#pragma once

#include "iterator2d.h"
#include <cassert>
#include <memory>
#include <type_traits>

namespace dim2d
{
#if defined(DEBUG)||defined(_DEBUG)
	const auto test = []( dim2d::index, Color ) { return true; };
#endif

	template<typename InIter, typename OutIter>
	void copy( InIter srcBeg, InIter end, OutIter dstBeg )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;
		
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);
		constexpr bool is_random_access_output =
			std::is_same_v<typename OutIter::iterator_category, std::random_access_iterator_tag>;
		std::is_standard_layout<OutIter>::value;
		auto it2 = dstBeg;
		const auto end2 = dstBeg + ( end - srcBeg );
		
		if constexpr( 
			std::is_trivially_copyable_v<type_1> && is_random_access_output )
		{
			const auto diff = end - srcBeg;
			const auto byteWidth = sizeof( type_2 ) * srcBeg.Width();

			for( auto it1 = srcBeg; it1 != end; it1 += offset{ 0,1 }, it2 += offset{ 0,1 } )
				memcpy_s( std::addressof( *it2 ), byteWidth, std::addressof( *it1 ), byteWidth );
		}
		else
		{
			for( auto it1 = srcBeg; it1 != end; ++it1, ++it2 )
				*it2 = *it1;
		}
	}

	template<typename InIter, typename OutIter, typename Pred>
	void copy_if( InIter src, InIter end, OutIter dst, Pred pred )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;

		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		for( auto it = src; it != end; ++it, ++dst )
			if( pred( it.Index(), *it ) )
				*dst = *it;
	}

	template<typename Iter, typename Func>
	void for_each( Iter beg, Iter end, Func func )
	{
		for( auto it = beg; it != end; ++it )
			func( it.Index(), *it );
	}

	template<typename Iter, typename T>
	void fill( Iter beg, Iter end, const T& val )
	{
		static_assert( std::is_same_v<typename Iter::value_type, std::decay_t<T> >,
			"Iterator::value_type and T don't match." );

		for( auto it = beg; it != end; ++it )
			*it = val;
	}

	template<typename InIter, typename OutIter, typename Func>
	void generate( InIter beg1, InIter end, OutIter beg2, Func gen )
	{
		using type_1  = typename InIter::value_type;
		using type_2  = typename OutIter::value_type;

		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		for( auto it = beg1; it != end; ++it, ++beg2 )
			*beg2 = gen( it.Index(), *it );
	}

	template<typename InIter, typename OutIter, typename T>
	void replace( InIter beg1, InIter end, OutIter beg2, const T& _this, const T& _that )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;

		static_assert( 
			std::conjunction_v<
			std::is_same<type_1, type_2>,
			std::is_same<type_1, T>>,
			"Container::value_type must be the same for both containers and T."
			);
		
		for( auto it = beg1; it != end; ++it, ++beg2 )
			if( ( *it ) == _this )
				*beg2 = _that;
	}

	template<typename InIter, typename OutIter, typename T, typename Pred>
	void replace_if( InIter beg1, InIter end, OutIter beg2, const T& val, Pred pred )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;

		static_assert(
			std::conjunction_v<
			std::is_same<type_1, type_2>,
			std::is_same<type_1, T>>,
			"Container::value_type must be the same for both containers and T."
			);

		for( auto it = beg1; it != end; ++it, ++beg2 )
		{
			if( pred( it.Index(), *it ) )
			{
				*beg2 = val;
			}
		}
	}

	template<typename InIter, typename OutIter, typename Func>
	void transform( InIter beg1, InIter end, OutIter beg2, Func func )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;

		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers and T."
			);

		for( auto it = beg1; it != end; ++it, ++beg2 )
			*beg2 = func( it.Index(), *it );
	}

	template<typename InIter, typename OutIter, typename Func>
	void transform( InIter beg1, InIter end, InIter beg2, OutIter beg3, Func func )
	{
		using type_1 = typename InIter::value_type;
		using type_2 = typename OutIter::value_type;

		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers and T."
			);

		for( auto it1 = beg1, it2 = beg2; it1 != end; ++it1, ++it2, ++beg3 )
			*beg3 = func( it1.Index(), *it1, it2.Index(), *it2 );
	}

#pragma region RECT_ALGORITHMS
	template<typename Cont1, typename Cont2, typename RectType>
	void copy( const RectType& _bounds, const Cont1& _source, Cont2& _dest )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		const auto width = _bounds.width();
		const auto stride = width * sizeof( type_1 );
		for( int y = _bounds.top(); y < _bounds.bottom(); ++y )
		{
			if constexpr( std::is_trivially_copyable_v<type_1> )
			{
				memcpy( std::addressof( _dest( 0, y ) ), std::addressof( _source( 0, y ) ), stride );
			}
			else
			{
				for( int x = _bounds.left(); x < _bounds.right(); ++x )
				{
					_dest( x, y ) = _source( x, y );
				}
			}
		}
	}

	template<typename Cont1, typename Cont2, typename RectType>
	void copy( const RectType& _srcBounds, const RectType& _dstBounds, const Cont1& _source, Cont2& _dest )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		assert(
			_srcBounds.Width() == _dstBounds.Width() &&
			_srcBounds.Height() == _dstBounds.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		const auto stride = sizeof( type_1 ) * _srcBounds.Width();
		for( int y = _srcBounds.Top(); y < _srcBounds.Bottom(); ++y )
		{
			const auto srcy = y + _srcBounds.Top();
			const auto dsty = y + _dstBounds.Top();
			if constexpr( std::is_trivially_copyable_v<type_1> )
			{
				auto* dst = std::addressof( _dest( _dstBounds.Left(), dsty ) );
				auto const* src = std::addressof( _source( _srcBounds.Left(), srcy ) );
				memcpy( dst, src, stride );
			}
			else
			{
				for( int x = _srcBounds.Left(); x < _srcBounds.Right(); ++x )
				{
					for( int x = 0; x < _srcBounds.Width(); ++x )
					{
						const auto srcx = x + _srcBounds.Left();
						const auto dstx = x + _dstBounds.Left();

						_dest( dstx, dsty ) = _source( srcx, srcy );
					}
				}
			}
		}
	}

	template<typename Cont1, typename Cont2, typename RectType, typename Pred>
	void copy_if(
		const RectType& _srcBounds, const RectType& _dstBounds, const Cont1& _source, Cont2& _dest,
		Pred&& _pred )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		assert(
			_srcBounds.Width() == _dstBounds.Width() &&
			_srcBounds.Height() == _dstBounds.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		for( int y = 0; y < _srcBounds.Height(); ++y )
		{
			const auto srcy = y + _srcBounds.Top();
			const auto dsty = y + _dstBounds.Top();

			for( int x = 0; x < _srcBounds.Width(); ++x )
			{
				const auto srcx = x + _srcBounds.Left();
				const auto dstx = x + _dstBounds.Left();

				if( const auto& elem = _source( srcx, srcy ); _pred( elem ) )
					_dest( dstx, dsty ) = elem;
			}
		}
	}

	template<typename RectType, typename Cont, typename Func>
	void for_each( const RectType& _bounds, Cont& _container, Func&& _func )
	{
		for( std::int32_t y = _bounds.Top(); y < _bounds.Bottom(); ++y )
		{
			for( std::int32_t x = _bounds.Left(); x < _bounds.Right(); ++x )
			{
				_func( _container( x, y ) );
			}
		}
	}

	template<typename RectType, typename Cont>
	void fill( const typename Cont::value_type& _value, const RectType& _bounds, Cont& _container )
	{
		for( int y = _bounds.Top(); y < _bounds.Bottom(); ++y )
		{
			for( int x = _bounds.Left(); x < _bounds.Right(); ++x )
			{
				_container( x, y ) = _value;
			}
		}
	}

	template<typename RectType, typename Cont, typename Generator>
	void generate( const RectType& _bounds, Cont& _cont, Generator&& _generate )
	{
		for( int y = _bounds.Top(); y < _bounds.Bottom(); ++y )
		{
			for( int x = _bounds.Left(); x < _bounds.Right(); ++x )
			{
				_cont( x, y ) = _generate();
			}
		}
	}

	template<typename RectType, typename Cont1, typename Cont2>
	void replace(
		const typename Cont1::value_type& _replace_this,
		const typename Cont2::value_type& _with_that,
		const RectType& _srcBounds,
		const RectType& _dstBounds,
		const Cont1& _src,
		Cont2& _dst )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		assert(
			_srcBounds.Width() == _dstBounds.Width() &&
			_srcBounds.Height() == _dstBounds.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		for( int y = 0; y < _srcBounds.height; ++y )
		{
			const auto srcy = y + _srcBounds.Top();
			const auto dsty = y + _dstBounds.Top();

			for( int x = 0; x < _srcBounds.width; ++x )
			{
				const auto srcx = x + _srcBounds.Left();
				const auto dstx = x + _dstBounds.Left();

				if( _src( srcx, srcy ) == _replace_this )
					_dst( dstx, dsty ) = _with_that;
			}
		}
	}

	template<typename RectType, typename Cont1, typename Cont2, typename Pred>
	void replace_if(
		const typename Cont1::value_type& _new_value,
		const RectType& _srcBounds,
		const RectType& _dstBounds,
		const Cont1& _src,
		Cont2& _dst,
		Pred&& _pred )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		assert(
			_srcBounds.Width() == _dstBounds.Width() &&
			_srcBounds.Height() == _dstBounds.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		for( int y = 0; y < _srcBounds.height; ++y )
		{
			const auto srcy = y + _srcBounds.Top();
			const auto dsty = y + _dstBounds.Top();

			for( int x = 0; x < _srcBounds.width; ++x )
			{
				const auto srcx = x + _srcBounds.Left();
				const auto dstx = x + _dstBounds.Left();

				const auto& srcElement = _src( srcx, srcy );
				auto& dstElement = _dst( dstx, dsty );

				if( _pred( srcElement, dstElement ) )
				{
					dstElement = _new_value;
				}
			}
		}
	}

	template<typename RectType, typename Cont1, typename Cont2, typename Func>
	void transform(
		const RectType& _srcBounds, const RectType& _dstBounds,
		const Cont1& _src, Cont2& _dst, Func&& _func )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		static_assert( std::is_same_v<type_1, type_2>,
			"Container::value_type must be the same for both containers."
			);

		assert(
			_srcBounds.Width() == _dstBounds.Width() &&
			_srcBounds.Height() == _dstBounds.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		for( int y = 0; y < _srcBounds.Height(); ++y )
		{
			const auto srcy = y + _srcBounds.Top();
			const auto dsty = y + _dstBounds.Top();

			for( int x = 0; x < _srcBounds.Width(); ++x )
			{
				const auto srcx = x + _srcBounds.Left();
				const auto dstx = x + _dstBounds.Left();

				_dst( dstx, dsty ) = _func( _src( srcx, srcy ) );
			}
		}
	}

	template<typename RectType, typename Cont1, typename Cont2, typename Cont3, typename Func>
	void transform(
		const RectType& _srcBounds1, const RectType& _srcBounds2, const RectType& _dstBounds,
		const Cont1& _src1, const Cont2& _src2, Cont3& _dst, Func&& _func )
	{
		using type_1 = typename Cont1::value_type;
		using type_2 = typename Cont2::value_type;
		using type_3 = typename Cont3::value_type;
		static_assert( std::conjunction_v<std::is_same<type_1, type_2>, std::is_same<type_2, type_3>>,
			"Container::value_type must be the same for all three containers."
			);

		assert(
			_srcBounds1.Width() == _dstBounds.Width() &&
			_srcBounds1.Height() == _dstBounds.Height() &&
			_srcBounds2.Width() == _srcBounds1.Width() &&
			_srcBounds2.Height() == _srcBounds1.Height() &&
			"Source and destination RectTypes must be same dimensions." );

		for( int y = 0; y < _srcBounds1.Height(); ++y )
		{
			const auto srcy1 = y + _srcBounds1.Top();
			const auto srcy2 = y + _srcBounds2.Top();
			const auto dsty = y + _dstBounds.Top();

			for( int x = 0; x < _srcBounds1.Width(); ++x )
			{
				const auto srcx1 = x + _srcBounds1.Left();
				const auto srcx2 = x + _srcBounds2.Left();
				const auto dstx = x + _dstBounds.Left();

				_dst( dstx, dsty ) = _func( _src1( srcx1, srcy1 ), _src2( srcx2, srcy2 ) );
			}
		}
	}
}
#pragma endregion