#pragma once

#include <iterator>
#include <stdexcept>
#include <type_traits>

namespace dim2d
{
	struct offset
	{
		int x = 0, y = 0;
	};

	constexpr offset operator+( const offset _left, const offset _right )noexcept
	{
		return { _left.x + _right.x, _left.y + _right.y };
	}
	constexpr offset operator-( const offset _left, const offset _right )noexcept
	{
		return { _left.x - _right.x, _left.y - _right.y };
	}
	using difference = offset;
	using index = offset;
}

namespace dim2d::detail
{
	template<
		typename iter_cat,
		typename val_type,
		typename ref_type,
		typename ptr_type,
		typename diff_type,
		typename idx_type,
		typename iter_child>
		class index_base
	{
	public:
		using iterator_category = iter_cat;
		using value_type = val_type;
		using reference = ref_type;
		using pointer = ptr_type;
		using difference_type = diff_type;
		using index_type = idx_type;
		using iterator = iter_child;

	public:
		index_base() = default;
		constexpr index_base(
			index_type _left,
			index_type _top,
			index_type _right,
			index_type _bottom )noexcept
			:
			left( _left ), top( _top ),
			x( _left ),
			y( _top ),
			width( _right - _left ),
			height( _bottom - _top )
		{}

		constexpr iterator operator+( offset _offset )const noexcept
		{
			auto result = *this;
			result.x += _offset.x;
			result.y += _offset.y;
			result.wrap();

			return static_cast< iterator& >( result );
		}
		constexpr iterator operator-( offset _offset )const noexcept
		{
			auto result = *this;
			result.x -= _offset.x;
			result.y -= _offset.y;
			result.wrap();

			return static_cast< iterator& >( result );
		}

		constexpr iterator& operator+=( offset _offset )noexcept
		{
			x += _offset.x;
			y += _offset.y;
			wrap();

			return reinterpret_cast< iterator& >( *this );
		}
		constexpr iterator& operator-=( offset _offset )noexcept
		{
			x -= _offset.x;
			y -= _offset.y;
			wrap();

			return reinterpret_cast< iterator& >( *this );
		}

		constexpr iterator& operator++()noexcept
		{
			++x;
			wrap();

			return static_cast< iterator& >( *this );
		}
		constexpr iterator operator++( int )noexcept
		{
			auto old = *this;
			++( *this );
			return reinterpret_cast< iterator& >( old );
		}

		constexpr iterator& operator--()noexcept
		{
			--x;
			wrap();

			return static_cast< iterator& >( *this );
		}
		constexpr iterator operator--( int )noexcept
		{
			auto old = *this;
			--( *this );
			return reinterpret_cast< iterator& >( old );
		}

		constexpr difference operator-( iterator _other )const noexcept
		{
			return { x - _other.x, y - _other.y };
		}

		constexpr bool operator==( iterator _other )const noexcept
		{
			return ( x == _other.x ) && ( y == _other.y );
		}
		constexpr bool operator!=( iterator _other )const noexcept
		{
			return !( ( *this ) == _other );
		}

		constexpr index_type Left()const noexcept
		{
			return left;
		}
		constexpr index_type Top()const noexcept
		{
			return top;
		}
		constexpr index_type Width()const noexcept
		{
			return width;
		}
		constexpr index_type Height()const noexcept
		{
			return height;
		}
		constexpr index_type X()const noexcept
		{
			return x;
		}
		constexpr index_type Y()const noexcept
		{
			return y;
		}
		constexpr index Index()const noexcept
		{
			return { x, y };
		}

	private:
		void wrap()
		{
			const auto right = left + width;
			if( x < left )
			{
				x = right - 1;
				--y;
			}
			else if( x >= right )
			{
				x = left;
				++y;
			}
		}
	protected:
		index_type left = 0, top = 0;
		index_type x = 0, y = 0;
		index_type width = 0, height = 0;
	};
}

namespace dim2d
{
	template<typename Container>
	class index_iterator : public detail::index_base<
		std::random_access_iterator_tag,
		typename Container::value_type,
		typename Container::reference,
		typename Container::pointer,
		difference,
		std::int32_t,
		index_iterator<Container>>
	{
	public:
		using base = detail::index_base<
			std::random_access_iterator_tag,
			typename Container::value_type,
			typename Container::reference,
			typename Container::pointer,
			difference,
			std::int32_t,
			index_iterator<Container>>;

		// STD iterator traits
		using iterator_category		= typename base::iterator_category;
		using value_type			= typename base::value_type;
		using reference				= typename base::reference;
		using pointer				= typename base::pointer;
		using difference_type		= typename base::difference_type;
		using index_type			= typename base::index_type;

		// DIM2D iterator traits
		using container_type		= Container;
		using container_reference	= container_type&;
		using container_pointer		= container_type*;

	public:
		index_iterator() = default;
		constexpr index_iterator(
			index_type _left,
			index_type _top,
			index_type _right,
			index_type _bottom,
			Container& _cont )noexcept
			:
			base( _left, _top, _right, _bottom), 
			cont( std::addressof( _cont ) )
		{}

		constexpr index_iterator& operator()( int _x, int _y )noexcept
		{
			x = _x;
			y = _y;

			return *this;
		}

		constexpr reference operator*() noexcept
		{
			return ( *cont )[ offset{ x,y } ];
		}

		constexpr pointer operator->()noexcept
		{
			return std::addressof( ( *cont )[ offset{ x,y } ] );
		}

	private:
		container_pointer cont = nullptr;
	};

	template<typename Container>
	class const_index_iterator : public detail::index_base<
		std::random_access_iterator_tag,
		typename Container::value_type,
		typename Container::const_reference,
		typename Container::const_pointer,
		difference,
		std::int32_t,
		const_index_iterator<Container>>
	{
	public:
		using base = detail::index_base<
			std::random_access_iterator_tag,
			typename Container::value_type,
			typename Container::const_reference,
			typename Container::const_pointer,
			difference,
			std::int32_t,
			const_index_iterator<Container>>;

		using iterator_category		= typename base::iterator_category;
		using value_type			= typename base::value_type;
		using reference				= typename base::reference;
		using pointer				= typename base::pointer;
		using difference_type		= typename base::difference_type;
		using index_type			= typename base::index_type;
		using container_type		= Container;
		using container_reference	= container_type const&;
		using container_pointer		= container_type const*;

	public:
		const_index_iterator() = default;
		constexpr const_index_iterator( 
			index_type _left,
			index_type _top,
			index_type _right,
			index_type _bottom, 
			Container const& _cont )noexcept
			:
			base( _left, _top, _right, _bottom),
			cont( std::addressof( _cont ) )
		{}

		constexpr const_index_iterator& operator()( int _x, int _y )noexcept
		{
			x = _x;
			y = _y;
			
			return *this;
		}

		constexpr reference operator*()const noexcept
		{
			return ( *cont )[ offset{ x,y } ];
		}

		constexpr pointer operator->()const noexcept
		{
			return std::addressof( ( *cont )[ offset{ x,y } ] );
		}

	private:
		container_pointer cont = nullptr;
	};

	template<typename IndexIter>
	class reverse_index_iterator
	{
	public:
		// STL traits
		using iterator_category = std::bidirectional_iterator_tag;
		using value_type = typename std::iterator_traits<IndexIter>::value_type;
		using reference = typename std::iterator_traits<IndexIter>::reference;
		using pointer = typename std::iterator_traits<IndexIter>::pointer;

		// DIM2D traits
		using index_type = typename IndexIter::index_type;
		using container_type = typename IndexIter::container_type;
		using container_reference = typename IndexIter::container_reference;
		using container_pointer = typename IndexIter::container_pointer;
		using iterator = reverse_index_iterator<IndexIter>;
		using wrapped_type = IndexIter;

	public:
		reverse_index_iterator() = default;
		constexpr reverse_index_iterator( wrapped_type _wrapped_iter )
			:
			wrapped( _wrapped_iter )
		{
			
		}
		constexpr reverse_index_iterator(
			index_type _left,
			index_type _top,
			index_type _right,
			index_type _bottom,
			container_reference _cont )noexcept
			:
			wrapped( _left, _top, _right, _bottom, _cont )
		{}
		constexpr iterator operator+( offset _offset )const noexcept
		{
			return iterator( *this ).wrapped( wrapped.X() - _offset.x, wrapped.Y() - _offset.y );
		}
		constexpr iterator operator-( offset _offset )const noexcept
		{
			return iterator( *this ).wrapped( wrapped.X() + _offset.x, wrapped.Y() + _offset.y );
		}

		constexpr iterator& operator+=( offset _offset )noexcept
		{
			wrapped( wrapped.X() - _offset.x, wrapped.Y() - _offset.y );
			wrap();

			return *this;
		}
		constexpr iterator& operator-=( offset _offset )noexcept
		{
			wrapped( wrapped.X() + _offset.x, wrapped.Y() + _offset.y );
			wrap();

			return *this;
		}

		constexpr iterator& operator++()noexcept
		{
			wrapped( wrapped.X() - 1, wrapped.Y() );
			wrap();

			return *this;
		}
		constexpr iterator operator++( int )noexcept
		{
			auto old = *this;
			++( *this );
			return old;
		}

		constexpr iterator& operator--()noexcept
		{
			wrapped( wrapped.X() + 1, wrapped.Y() );
			wrap();

			return *this;
		}
		constexpr iterator operator--( int )noexcept
		{
			auto old = *this;
			--( *this );
			return old;
		}

		constexpr difference operator-( iterator const& _other )const noexcept
		{
			return _other.wrapped - wrapped;
		}

		constexpr bool operator==( iterator const& _other )const noexcept
		{
			return  wrapped == _other.wrapped;
		}
		constexpr bool operator!=( iterator const& _other )const noexcept
		{
			return !( ( *this ) == _other );
		}

		constexpr index_type Left()const noexcept
		{
			return wrapped.Left();
		}
		constexpr index_type Top()const noexcept
		{
			return wrapped.Top();
		}
		constexpr index_type Width()const noexcept
		{
			return wrapped.Width();
		}
		constexpr index_type X()const noexcept
		{
			return wrapped.X();
		}
		constexpr index_type Y()const noexcept
		{
			return wrapped.Y();
		}
		constexpr index Index()const noexcept
		{
			return wrapped.Index();
		}
		
		constexpr auto operator*()noexcept
			->std::enable_if_t<std::negation_v<std::is_const<reference>>, value_type&>
		{
			auto temp = *this;
			--temp;
			return *temp.wrapped;
		}
		constexpr auto operator*()const noexcept
			->std::enable_if_t<std::is_const_v<reference>, value_type const&>
		{
			auto temp = *this;
			--temp;
			return *temp.wrapped;
		}


		constexpr auto operator->()noexcept
			->std::enable_if_t<std::negation_v<std::is_const<pointer>>, value_type*>
		{
			auto temp = *this;
			++temp;
			return temp.wrapped.operator->();
		}

		constexpr auto operator->()const noexcept
			->std::enable_if_t<std::is_const_v<pointer>, value_type*>
		{
			auto temp = *this;
			++temp;
			return temp.wrapped.operator->();
		}
	private:
		void wrap()
		{
			const auto right = Left() + Width();
			if( X() < Left() )
			{
				wrapped( right - 1, Y() - 1 );
			}
			else if( X() >= right )
			{
				wrapped( Left(), Y() + 1 );
			}
		}
	protected:
		wrapped_type wrapped;
	};

	template<typename IndexIter>
	class mirror_index_iterator
	{
	public:
		// STL traits
		using iterator_category = std::bidirectional_iterator_tag;
		using value_type = typename std::iterator_traits<IndexIter>::value_type;
		using reference = typename std::iterator_traits<IndexIter>::reference;
		using pointer = typename std::iterator_traits<IndexIter>::pointer;

		// DIM2D traits
		using index_type = typename IndexIter::index_type;
		using container_type = typename IndexIter::container_type;
		using container_reference = typename IndexIter::container_reference;
		using container_pointer = typename IndexIter::container_pointer;
		using iterator = mirror_index_iterator<IndexIter>;
		using wrapped_type = IndexIter;

	public:
		mirror_index_iterator() = default;
		constexpr mirror_index_iterator( wrapped_type _wrapped_iter )
			:
			wrapped( _wrapped_iter )
		{

		}
		constexpr mirror_index_iterator(
			index_type _left,
			index_type _top,
			index_type _right,
			index_type _bottom,
			container_reference _cont )noexcept
			:
			wrapped( _left, _top, _right, _bottom, _cont )			
		{
			wrapped( _left, _bottom );
		}
		constexpr iterator operator+( offset _offset )const noexcept
		{
			return iterator( *this ).wrapped( wrapped.X() - _offset.x, wrapped.Y() + _offset.y );
		}
		constexpr iterator operator-( offset _offset )const noexcept
		{
			return iterator( *this ).wrapped( wrapped.X() + _offset.x, wrapped.Y() - _offset.y );
		}

		constexpr iterator& operator+=( offset _offset )noexcept
		{
			wrapped( wrapped.X() - _offset.x, wrapped.Y() + _offset.y );
			wrap();

			return *this;
		}
		constexpr iterator& operator-=( offset _offset )noexcept
		{
			wrapped( wrapped.X() + _offset.x, wrapped.Y() - _offset.y );
			wrap();

			return *this;
		}

		constexpr iterator& operator++()noexcept
		{
			wrapped( wrapped.X() - 1, wrapped.Y() );
			wrap();

			return *this;
		}
		constexpr iterator operator++( int )noexcept
		{
			auto old = *this;
			++( *this );
			return old;
		}

		constexpr iterator& operator--()noexcept
		{
			wrapped( wrapped.X() + 1, wrapped.Y() );
			wrap();

			return *this;
		}
		constexpr iterator operator--( int )noexcept
		{
			auto old = *this;
			--( *this );
			return old;
		}

		constexpr difference operator-( iterator const& _other )const noexcept
		{
			return _other.wrapped - wrapped;
		}

		constexpr bool operator==( iterator const& _other )const noexcept
		{
			return  wrapped == _other.wrapped;
		}
		constexpr bool operator!=( iterator const& _other )const noexcept
		{
			return !( ( *this ) == _other );
		}

		constexpr index_type Left()const noexcept
		{
			return wrapped.Left();
		}
		constexpr index_type Top()const noexcept
		{
			return wrapped.Top();
		}
		constexpr index_type Width()const noexcept
		{
			return wrapped.Width();
		}
		constexpr index_type X()const noexcept
		{
			return wrapped.X();
		}
		constexpr index_type Y()const noexcept
		{
			return wrapped.Y();
		}
		constexpr index Index()const noexcept
		{
			return wrapped.Index();
		}

		constexpr auto operator*()noexcept
			->std::enable_if_t<std::negation_v<std::is_const<reference>>, value_type&>
		{
			auto temp = *this;
			++temp;
			return *temp.wrapped;
		}
		constexpr auto operator*()const noexcept
			->std::enable_if_t<std::is_const_v<reference>, value_type const&>
		{
			auto temp = *this;
			++temp;
			return *temp.wrapped;
		}


		constexpr auto operator->()noexcept
			->std::enable_if_t<std::negation_v<std::is_const<pointer>>, value_type*>
		{
			auto temp = *this;
			++temp;
			return temp.wrapped.operator->();
		}

		constexpr auto operator->()const noexcept
			->std::enable_if_t<std::is_const_v<pointer>, value_type*>
		{
			auto temp = *this;
			++temp;
			return temp.wrapped.operator->();
		}
	private:
		void wrap()
		{
			const auto right = Left() + Width();
			if( X() < Left() )
			{
				wrapped( right - 1, Y() + 1 );
			}
			else if( X() >= right )
			{
				wrapped( Left(), Y() - 1 );
			}
		}
	protected:
		wrapped_type wrapped;
	};

	template<typename Container>
	index_iterator( int32_t, int32_t, int32_t, int32_t, Container& )
		->index_iterator<Container>;

	template<typename Container>
	const_index_iterator( int32_t, int32_t, int32_t, int32_t, Container const& )
		->const_index_iterator<Container>;
}
