#pragma once

#include "iterator2d.h"
#include <cassert>
#include <vector>

//#include "../chili_framework/Engine/Sprite.h"

namespace dim2d
{
	template<typename T>
	class surface
	{
	public:
		using value_type				= T;
		using reference					= T & ;
		using const_reference			= T const&;
		using pointer					= T * ;
		using const_pointer				= T const*;
		using iterator					= dim2d::index_iterator<surface<T>>;
		using const_iterator			= dim2d::const_index_iterator<surface<T>>;
		using reverse_iterator			= dim2d::reverse_index_iterator<iterator>;
		using const_reverse_iterator	= dim2d::reverse_index_iterator<const_iterator>;

	public:
		surface() = default;
		surface( int _width, int _height )
			:
			width( _width ),
			height( _height ),
			ptr( new T[ _width * _height ] )
		{}
		surface( surface const& ) = delete;
		surface( surface&& _other )
		{
			*this = std::move( _other );
		}
		~surface()
		{
			if( ptr )
			{
				delete[] ptr;
				ptr = nullptr;
				width = 0;
				height = 0;
			}
		}
		surface& operator=( surface const& _other ) = delete;
		surface& operator=( surface&& _other )
		{
			if( this != std::addressof( _other ) )
			{
				ptr = _other.ptr;
				width = _other.width;
				height = _other.height;

				_other.ptr = nullptr;
				_other.width = 0;
				_other.height = 0;
			}

			return *this;
		}

		iterator begin()noexcept 
		{ 
			return iterator( 0, 0, width, height, *this ); 
		}
		iterator end()noexcept 
		{ 
			return iterator( 0, height, width, height, *this ); 
		}

		const_iterator begin()const noexcept 
		{ 
			return const_iterator( 0, 0, width, height, *this ); 
		}
		const_iterator end()const noexcept 
		{ 
			return const_iterator( 0, height, width, height, *this ); 
		}

		const_iterator cbegin()const noexcept 
		{ 
			begin(); 
		}
		const_iterator cend()const noexcept 
		{ 
			end(); 
		}

		reverse_iterator rbegin()noexcept 
		{ 
			end(); 
		}
		reverse_iterator rend()noexcept 
		{ 
			begin(); 
		}

		reverse_iterator crbegin()const noexcept 
		{ 
			end() - offset{ 1,0 }; 
		}
		reverse_iterator rend()const noexcept 
		{ 
			begin() - offset{ 1,0 }; 
		}

		constexpr int size()const noexcept 
		{ 
			return width * height; 
		}
		constexpr int rows()const noexcept 
		{ 
			return height; 
		}
		constexpr int columns()const noexcept 
		{ 
			return width; 
		}

		constexpr reference operator[]( offset _idx )noexcept 
		{
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );
			return ptr[ _idx.x + ( _idx.y * width ) ]; 
		}
		constexpr const_reference operator[](offset _idx)const noexcept
		{ 
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );
			return ptr[ _idx.x + ( _idx.y * width ) ];
		}

	private:
		T* ptr = nullptr;
		int width = 0, height = 0;
	};

	template<typename T>
	class raw_pointer_wrapper
	{
	public:
		using value_type = T;
		using reference = T & ;
		using const_reference = const T&;
		using pointer = T * ;
		using const_pointer = const T*;
		using iterator = dim2d::index_iterator<raw_pointer_wrapper<T>>;
		using const_iterator = dim2d::const_index_iterator<raw_pointer_wrapper<T>>;
		using storage = T * ;

	public:
		raw_pointer_wrapper() = default;
		raw_pointer_wrapper( offset _offset, int _width, int _height, int _stride, T* _source )
			:
			width( _width ),
			height( _height ),
			stride( _stride ),
			source( _source )
		{
		}
		raw_pointer_wrapper( raw_pointer_wrapper const& ) = delete;
		raw_pointer_wrapper( raw_pointer_wrapper&& ) = delete;
		~raw_pointer_wrapper() {
			if constexpr( std::is_pointer_v<T> )
			{
				if( source ) {
					source = nullptr;
					width = 0;
					height = 0;
					stride = 0;
					m_offset = { 0, 0 };
				}
			}
		}
		raw_pointer_wrapper& operator=( raw_pointer_wrapper const& _other ) = delete;
		raw_pointer_wrapper& operator=( raw_pointer_wrapper&& _other ) = delete;

		iterator begin()noexcept {
			return iterator( 0, 0, width, height, *this );
		}
		iterator end()noexcept {
			return iterator( 0, height, width, height, *this );
		}

		const_iterator begin()const noexcept {
			return const_iterator( 0, 0, width, height, *this );
		}
		const_iterator end()const noexcept {
			return const_iterator( 0, height, width, height, *this );
		}

		const_iterator cbegin()const noexcept {
			return begin();
		}
		const_iterator cend()const noexcept {
			return end();
		}

		constexpr int size()const noexcept {
			return width * height;
		}
		constexpr int rows()const noexcept {
			return height;
		}
		constexpr int columns()const noexcept {
			return width;
		}

		constexpr reference operator[]( offset _idx )noexcept {
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );

			const auto idx_offset = _idx + m_offset;
			const auto idx_1D = idx_offset.x + ( idx_offset.y * stride );
			return source[ idx_1D ];
		}
		constexpr const_reference operator[]( offset _idx )const noexcept {
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );

			const auto idx_offset = _idx + m_offset;
			const auto idx_1D = idx_offset.x + ( idx_offset.y * stride );
			return source[ idx_1D ];
		}

	private:
		storage source;
		int width = 0, height = 0, stride = 0;
		offset m_offset;
	};

	template<typename T>
	class surface_wrapper 
	{
	public:
		using value_type = typename std::remove_reference_t<T>::value_type;
		using pointer = value_type * ;
		using const_pointer = value_type const*;
		using reference = value_type & ;
		using const_reference = value_type const&;
		using iterator = dim2d::index_iterator<surface_wrapper<T>>;
		using const_iterator = dim2d::const_index_iterator<surface_wrapper<T>>;
		using storage = std::conditional_t<std::is_const_v<T>, T const&, T &>;;

		surface_wrapper() = default;
		surface_wrapper( offset _offset, int _width, int _height, int _stride, storage _source )
			:
			m_offset( _offset ),
			width( _width ),
			height( _height ),
			stride( _stride ),
			source( _source )
		{
		}
		surface_wrapper( surface_wrapper const& ) = delete;
		surface_wrapper( surface_wrapper&& ) = delete;

		surface_wrapper& operator=( surface_wrapper const& _other ) = delete;
		surface_wrapper& operator=( surface_wrapper&& _other ) = delete;

		iterator begin()noexcept {
			return iterator( 0, 0, width, height, *this );
		}
		iterator end()noexcept {
			return iterator( 0, height, width, height, *this );
		}

		const_iterator begin()const noexcept {
			return const_iterator( 0, 0, width, height, *this );
		}
		const_iterator end()const noexcept {
			return const_iterator( 0, height, width, height, *this );
		}

		const_iterator cbegin()const noexcept {
			return begin();
		}
		const_iterator cend()const noexcept {
			return end();
		}

		constexpr int size()const noexcept {
			return width * height;
		}
		constexpr int rows()const noexcept {
			return height;
		}
		constexpr int columns()const noexcept {
			return width;
		}

		constexpr reference operator[]( offset _idx )noexcept
		{
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );

			return source[ _idx + m_offset ];
		}
		constexpr const_reference operator[]( offset _idx )const noexcept
		{
			assert( _idx.x >= 0 );
			assert( _idx.x < width );
			assert( _idx.y >= 0 );
			assert( _idx.y < height );

			return source[ _idx + m_offset ];
		}

	protected:
		storage source;
		int width = 0, height = 0, stride = 0;
		offset m_offset;
	};

	// Template deduction guides
	/*template<typename T> 
	surface_wrapper( offset, int, int, int, T& )->surface_wrapper<T>;
	template<typename T>
	surface_wrapper( offset, int, int, int, T const& )->surface_wrapper<T>;*/
	template<typename T>
	surface_wrapper( offset, int, int, int, T& )->surface_wrapper<T>;

	template<typename T>
	raw_pointer_wrapper( offset, int, int, int, T* )->raw_pointer_wrapper<T>;
}